/* This file is part of SWIS
   Copyright (C) 2007 Sergey Poznyakoff
 
   SWIS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SWIS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SWIS.  If not, see <http://www.gnu.org/licenses/>. */

#include "swis.h"
#include <stdarg.h>
#include <fcntl.h>
#include <mysql/mysql.h>
#include "getpass.h"

enum {
  PROGNAME_OPTION,
  CLEANUP_OPTION,
  PASS_FROM_FILE_OPTION,
  PASS_FROM_FD_OPTION,
  URL_TABLE_OPTION,
  WORD_TABLE_OPTION,
  TEXT_TABLE_OPTION
};

struct option options[] = {
  { "progname", required_argument, NULL, PROGNAME_OPTION },
  { "file", required_argument, NULL, 'f' },
  { "debug", no_argument, NULL, 'd' },
  { "cleanup", no_argument, NULL, CLEANUP_OPTION },
  { "user", required_argument, NULL, 'u' },
  { "password", optional_argument, NULL, 'p' },
  { "pass-from-file", required_argument, NULL, PASS_FROM_FILE_OPTION },
  { "pass-from-fd", required_argument, NULL, PASS_FROM_FD_OPTION },
  { "port", required_argument, NULL, 'P' },
  { "socket", required_argument, NULL, 'S' },
  { "host", required_argument, NULL, 'H' },
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { NULL }
};

void
usage ()
{
  printf ("Usage: word-split [OPTIONS] [DBNAME]\n");
  printf ("Read input and store it in a MySQL database\n");
  printf ("\n");
  printf ("DBNAME is the name of the MySQL database to use\n");
  printf ("\nOptions are:\n");
  printf ("  -d, --debug                    output debugging info\n"); 
  printf ("      --cleanup                  clean the database up before proceeding\n");
  printf ("  -f, --file=FILE                read from FILE instead of stdin\n");
  printf ("  -H, --host=HOST                MySQL host name\n");
  printf ("  -u, --user=USER                MySQL user name\n");
  printf ("  -p, --password[=PASS]          MySQL password\n");
  printf ("      --url-table=DESCR          describe url table\n");
  printf ("      --word-table=DESCR         describe word table\n");
  printf ("      --text-table=DESCR         describe text table\n");
  printf ("      --pass-from-file=FILE      read MySQL password from the given FILE\n");
  printf ("      --pass-from-fd=NUMBER      read MySQL password from the file descriptor\n");
  printf ("                                 NUMBER\n");
  printf ("  -P, --port=NUMBER              MySQL port number\n");
  printf ("  -S, --socket=FILE              MySQL socket name\n");
  
  printf ("  -h, --help                     print this help list\n");
  printf ("  -v, --version                  print program version and exit\n");
  printf ("\n");
  printf ("Table description argumend (DESCR) syntax is:\n");
  printf ("[name:]column=name[,column=name...].\n");
  printf ("\n");
  printf ("Report bugs to <%s>\n", PACKAGE_BUGREPORT);
}

FILE *input;
int debug_level;

/* MySQL access parameters */ 
char *host;                 /* host name */
char *socket_name;          /* socket ... */
int port;                   /* ... or port */
char *db = "swis";          /* Database name */
char *user;                 /* User name */
char *pass;                 /* Password */

int cleanup_option;

/* Table and field names */
#define COLTYPE_STRING  0
#define COLTYPE_NUMERIC 1
struct table_info
{
  char *name;
  int ncolumns;
  char **col_id;
  char **col_name;
  int *col_type;
};

int common_types[] = { COLTYPE_STRING, COLTYPE_NUMERIC };
enum { url_column_url, url_column_id };
char *url_columns[] = { "url", "id" };
char *url_names[] = { "url", "id" };
struct table_info url_tab = { "url",
			      2,
			      url_columns,
			      url_names,
			      common_types };

int url_auto_increment = url_column_id;

enum { word_column_word, word_column_pos, word_column_off, word_column_id };
char *word_columns[] = { "word", "pos", "off", "url_id" };
char *word_names[] = { "word", "pos", "off", "url_id" };
int word_types[] = { COLTYPE_STRING, COLTYPE_NUMERIC, COLTYPE_NUMERIC,
		     COLTYPE_NUMERIC };
struct table_info word_tab = { "words",
			       4,
			       word_columns,
			       word_names,
			       word_types };

enum { text_column_text, text_column_id };
char *text_columns[] = { "text", "url_id" };
char *text_names[] = { "text", "url_id" };
struct table_info text_tab = { "text",
			       2,
			       text_columns,
			       text_names,
			       common_types };


void
read_password_from_fd (int fd)
{
  struct obstack stk;
  char c, *p;
  
  obstack_init (&stk);
  while (read (fd, &c, 1) == 1 && c != '\n')
    obstack_1grow (&stk, c);
  obstack_1grow (&stk, 0);
  p = obstack_finish (&stk);
  pass = xstrdup (p);
  obstack_free (&stk, 0);
}

void
read_password_from_file (const char *filename)
{
  int fd = open (filename, O_RDONLY);
  if (fd == -1)
    error (1, errno, "cannot open file %s", filename);
  read_password_from_fd (fd);
  close (fd);
}


int
find_column_id (struct table_info *ti, const char *cid)
{
  int i;
  for (i = 0; i < ti->ncolumns; i++)
    if (strcmp (ti->col_id[i], cid) == 0)
      return i;
  return -1;
}

int
find_column_name (struct table_info *ti, const char *name)
{
  int i;
  for (i = 0; i < ti->ncolumns; i++)
    if (strcmp (ti->col_name[i], name) == 0)
      return i;
  return -1;
}

void
read_table_info (struct table_info *ti, const char *arg)
{
  char *tmp = xstrdup (arg);
  char *spec;
  char *p = strchr (tmp, ':');

  if (p)
    {
      *p++ = 0;
      spec = p;
      ti->name = xstrdup (spec);
    }
  else
    spec = tmp;
  
  for (spec = strtok (spec, ","); spec; spec = strtok (NULL, ","))
    {
      int i;

      p = strchr (spec, '=');
      if (!p)
	error (1, 0, "invalid table specification: %s (near char %d)",
	       arg, spec-tmp);
      *p++ = 0;
      i = find_column_id (ti, spec);
      if (i == -1)
	error (1, 0, "invalid table specification: %s: no such column %s",
	       arg, spec);
      ti->col_name[i] = xstrdup (p);
    }
  free (tmp);
}


MYSQL mysql;
MYSQL_RES *result;

unsigned long next_url_id;
unsigned long url_id;
unsigned long word_pos;

/* Low-level SQL functions */

struct escape_slot
{
  char *ptr;
  size_t size;
};

struct escape_slot escape_slots[3];

char *
escape_arg (int slot, const char *arg)
{
  size_t arglen = strlen (arg);
  size_t maxlen = arglen * 2 + 1;
  struct escape_slot *ps = &escape_slots[slot];
  
  if (!ps->ptr)
    {
      ps->size = maxlen;
      ps->ptr = xmalloc (ps->size);
    }
  else if (ps->size < maxlen)
    {
      ps->size = maxlen;
      ps->ptr = x2realloc (ps->ptr, &ps->size);
    }

  mysql_real_escape_string (&mysql, ps->ptr, arg, arglen);
  return ps->ptr;
}

void
exec_sql (const char *fmt, ...)
{
  va_list ap;
  char *qbuf = NULL;
  
  va_start (ap, fmt);
  vasprintf (&qbuf, fmt, ap);
  va_end (ap);
  if (!qbuf)
    error (1, errno, "cannot format query");

  if (result)
    {
      mysql_free_result (result);
      result = NULL;
    }

  if (debug_level)
    printf ("DEBUG: Query: %s\n", qbuf);
  
  if (mysql_query (&mysql, qbuf))
    {
      error (0, 0, "query failed: %s", mysql_error (&mysql));
      error (1, 0, "the failed query was: %s", qbuf);
    }

  if (debug_level)
    {
      unsigned long x = mysql_affected_rows (&mysql);
      printf ("DEBUG: the query affected %lu rows\n", x);
    }
}

int
get_numeric_result (unsigned long *var)
{
  unsigned long n;
  MYSQL_ROW row;
  char *p;
  
  if (!result)
    {
      result = mysql_store_result (&mysql);
      if (!result)
	return 1;
    }
  row = mysql_fetch_row (result);
  if (row == NULL || row[0] == NULL)
    return 1;
  n = strtoul (row[0], &p, 10);
  if (*p)
    {
      error (0, 0, "cannot convert string to number (near %s)", p);
      return 1;
    }
  *var = n;
  return 0;
}

void
assert_insertion ()
{
  if (mysql_affected_rows (&mysql) == 0)
    error (0, 0, "query affected 0 rows");
}



/* Connect / disconnect */

void
disconnect ()
{
  mysql_close (&mysql);
}

void
connect_to_sql ()
{
  mysql_init (&mysql);
  if (!mysql_real_connect (&mysql, host, user, pass, db, port, socket_name, 0))
    error (1, 0, "failed to connect to database %s: error: %s",
	   db, mysql_error (&mysql));
  exec_sql ("SET NAMES utf8");
  atexit (disconnect);
}


/* Database verification */
void
verify_table_sanity (struct table_info *tab, int *pautoinc)
{
  unsigned int num_fields;
  unsigned int i;
  MYSQL_FIELD *field;
  MYSQL_RES *result;
  int *col_ok;
  int errcount = 0;
  
  col_ok = xcalloc (tab->ncolumns, sizeof col_ok[0]);
  
  result = mysql_list_fields (&mysql, tab->name, NULL);
  if (!result)
    error (1, 0, "cannot get field list for %s.%s: %s", db, tab->name,
	   mysql_error (&mysql));

  num_fields = mysql_num_fields (result);
  for (i = 0; i < num_fields; i++)
    {
      int n;

      field = mysql_fetch_field_direct(result, i);
      if (!field)
	error (1, 0, "cannot get description of field %d in table %s", i,
	       tab->name);

      n = find_column_name (tab, field->name);
      if (n == -1)
	continue;

      col_ok[n] = 1;

      switch (tab->col_type[n])
	{
	case COLTYPE_STRING:
	  switch (field->type)
	    {
	    case FIELD_TYPE_STRING:
	    case FIELD_TYPE_VAR_STRING:
	    case FIELD_TYPE_BLOB:
	      break;

	    default:
	      errcount++;
	      error (0, 0, "column %s.%s is not of string type",
		     tab->name, tab->col_name[n]);
	    }
	  break;

	case COLTYPE_NUMERIC:
	  if (!IS_NUM (field->type))
	    {
	      errcount++;
	      error (0, 0, "column %s.%s is not of numeric type",
		     tab->name, tab->col_name[n]);
	    }
	  if (pautoinc && n == *pautoinc)
	    *pautoinc = field->flags & AUTO_INCREMENT_FLAG;
	}
    }
  mysql_free_result (result);

  for (i = 0; i < tab->ncolumns; i++)
    {
      if (!col_ok[i])
	{
	  error (0, 0, "no column %s in table %s", tab->col_name[i],
		 tab->name);
	  errcount++;
	}
    }
  free (col_ok);
  if (errcount)
    exit (1);
}

void
verify_db_sanity ()
{
  verify_table_sanity (&url_tab, &url_auto_increment);
  verify_table_sanity (&word_tab, NULL);
  verify_table_sanity (&text_tab, NULL);
  if (!url_auto_increment)
    {
      exec_sql ("SELECT MAX(%s)+1 FROM %s",
		url_tab.col_name[url_column_id], url_tab.name);
      if (get_numeric_result (&next_url_id))
	error (1, 0, "cannot get maximum URL ID: %s", mysql_error (&mysql));
    }
}


/* High-level functions */

void
cleanup_db ()
{
  exec_sql ("DELETE FROM %s", url_tab.name);
  exec_sql ("DELETE FROM %s", word_tab.name);
  exec_sql ("DELETE FROM %s", text_tab.name);
}

void
change_url (const char *url)
{
  char *p;
  
  exec_sql ("SELECT %s FROM %s WHERE %s='%s'",
	    url_tab.col_name[url_column_id], url_tab.name,
	    url_tab.col_name[url_column_url],
	    p = escape_arg (0, url));
  if (get_numeric_result (&url_id))
    {
      if (url_auto_increment)
	{
	  exec_sql ("INSERT INTO %s (%s) VALUES ('%s')",
		    url_tab.name, url_tab.col_name[url_column_url], p);
	  exec_sql ("SELECT LAST_INSERT_ID()");
	  if (get_numeric_result (&url_id))
	    error (1, 0, "cannot get URL ID: %s", mysql_error (&mysql));
	}
      else
	{
	  url_id = next_url_id++;
	  exec_sql ("INSERT INTO %s (%s,%s) VALUES (%s,%ul)",
		    url_tab.name, url_tab.col_name[url_column_url],
		    url_tab.col_name[url_column_id],
		    p, url_id);
	}
    }
  word_pos = 0;
}

void
store_text (const char *text)
{
  exec_sql ("INSERT INTO %s (%s,%s) VALUES ('%s',%lu)",
	    text_tab.name, text_tab.col_name[text_column_text],
	    text_tab.col_name[text_column_id],
	    escape_arg (0, text), url_id);
  assert_insertion ();
}

void
store_word (char *word, unsigned long off)
{
  exec_sql ("INSERT INTO %s (%s,%s,%s,%s) VALUES ('%s',%lu,%lu,%lu)",
	    word_tab.name,
	    word_tab.col_name[word_column_word],
	    word_tab.col_name[word_column_pos],
	    word_tab.col_name[word_column_off],
	    word_tab.col_name[word_column_id],
	    escape_arg (0, word), word_pos, off, url_id);
  assert_insertion ();
  word_pos++;
}

int
parse_record (char *buf, char **pword, unsigned long *pnum)
{
  char *p;
  
  *pnum = strtoul (buf, &p, 10);
  if (*p != ' ')
    {
      error (0, 0, "malformed input: %s", buf);
      return 1;
    }
  *pword = p + 1;
  return 0;
}


/* Main loop */
void
read_loop ()
{
  char *buf = NULL;
  size_t size = 0;
  char *word;
  unsigned long num;
  
  while (getline (&buf, &size, input) > 0)
    {
      int len = strlen (buf);

      if (buf[len-1] == '\n')
	buf[len-1] = 0;
      
      if (buf[0] == '>')
	{
	  switch (buf[1])
	    {
	    case ' ':
	      change_url (buf + 2);
	      break;

	    case '*':
	      store_text (buf + 2);
	      break;

	    default:
	      error (1, 0, "unexpected input: %s", buf);
	    }
	}
      else if (parse_record (buf, &word, &num) == 0)
	store_word (word, num);
    }
  free (buf);
}


int
main (int argc, char **argv)
{
  int c;
  int fd;
  char *p;
  
  program_name = argv[0];
  while ((c = getopt_long (argc, argv, "df:H:hP:p::S:u:v", options, NULL))
	 != EOF)
    {
      switch (c)
	{
	case CLEANUP_OPTION:
	  cleanup_option = 1;
	  break;

	case 'd':
	  debug_level++;
	  break;
	  
	case 'f':
	  input = fopen (optarg, "r");
	  if (!input)
	    error (1, errno, "cannot open file %s", optarg);
	  break;

	case 'H':
	  host = optarg;
	  break;
	  
	case 'h':
	  usage ();
	  exit (0);

	case PASS_FROM_FD_OPTION:
	  fd = strtoul (optarg, &p, 10);
	  if (*p)
	    error (1, 0, "not a number: %s", optarg);
	  read_password_from_fd (fd);
	  break;
	    
	case PASS_FROM_FILE_OPTION:
	  read_password_from_file (optarg);
	  break;
	  
	case PROGNAME_OPTION:
	  program_name = optarg;
	  break;

	case 'P':
	  port = strtoul (optarg, &p, 10);
	  if (*p)
	    error (1, 0, "not a number: %s", optarg);
	  break;
	  
	case 'p':
	  if (optarg)
	    pass = optarg;
	  else
	    pass = xstrdup (getpass ("Password:"));
	  break;
	  
	case 'S':
	  socket_name = optarg;
	  break;
	  
	case 'u':
	  user = optarg;
	  break;
	  
	case 'v':
	  swis_version (stdout, "mysql-backend");
	  exit (0);

	case URL_TABLE_OPTION:
	  read_table_info (&url_tab, optarg);
	  break;
	  
	case WORD_TABLE_OPTION:
	  read_table_info (&word_tab, optarg);
	  break;
	  
	case TEXT_TABLE_OPTION:
	  read_table_info (&text_tab, optarg);
	  break;
	  
	default:
	  exit (1);
	}
    }

  argc -= optind;
  argv += optind;

  if (argc >= 1)
    {
      if (argc > 1)
	error (0, 0, "extra arguments ignored");
      db = argv[0];
    }
  
  if (!input)
    input = stdin;

  connect_to_sql ();
  verify_db_sanity ();

  if (cleanup_option)
    cleanup_db ();

  read_loop ();

  exit (0);
}
